# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from torch import matmul, cat
from torch.nn.functional import normalize

from ..models import BilinearModel
from ..utils import init_embedding


class RESCALModel(BilinearModel):
    """Implementation of RESCAL model detailed in 2011 paper by Nickel et al..
    In the original paper, optimization is done using Alternating Least Squares
    (ALS). Here we use iterative gradient descent optimization. This class
    inherits from the :class:`torchkge.models.interfaces.BilinearModel`
    interface. It then has its attributes as well.

    References
    ----------
    * Maximilian Nickel, Volker Tresp, and Hans-Peter Kriegel.
      `A Three-way Model for Collective Learning on Multi-relational Data.
      <https://dl.acm.org/citation.cfm?id=3104584>`_
      In Proceedings of the 28th International Conference on Machine Learning,
      2011.


    Parameters
    ----------
    emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    ent_emb: torch.nn.Embedding, shape: (n_ent, emb_dim)
        Embeddings of the entities, initialized with Xavier uniform
        distribution and then normalized.
    rel_mat: torch.nn.Embedding, shape: (n_rel, emb_dim x emb_dim)
        Matrices of the relations, initialized with Xavier uniform
        distribution.

    """

    def __init__(self, emb_dim, n_entities, n_relations):
        super().__init__(emb_dim, n_entities, n_relations)

        # initialize embedding objects
        self.ent_emb = init_embedding(self.n_ent, self.emb_dim)
        self.rel_mat = init_embedding(self.n_rel, self.emb_dim * self.emb_dim)

        # normalize the embeddings
        self.normalize_parameters()

    def scoring_function(self, h_idx, t_idx, r_idx):
        """Compute the scoring function for the triplets given as argument:
        :math:`h^T \\cdot M_r \\cdot t`. See referenced paper for more details
        on the score. See torchkge.models.interfaces.Models for more details
        on the API.

        """
        h = normalize(self.ent_emb(h_idx), p=2, dim=1)
        t = normalize(self.ent_emb(t_idx), p=2, dim=1)
        r = self.rel_mat(r_idx).view(-1, self.emb_dim, self.emb_dim)
        hr = matmul(h.view(-1, 1, self.emb_dim), r)
        return (hr.view(-1, self.emb_dim) * t).sum(dim=1)

    def normalize_parameters(self):
        """Normalize the entity embeddings, as explained in original paper.
        This methods should be called at the end of each training epoch and at
        the end of training as well.

        """
        self.ent_emb.weight.data = normalize(self.ent_emb.weight.data,
                                             p=2, dim=1)

    def get_embeddings(self):
        """Return the embeddings of entities and matrices of relations.

        Returns
        -------
        ent_emb: torch.Tensor, shape: (n_ent, emb_dim), dtype: torch.float
            Embeddings of entities.
        rel_mat: torch.Tensor, shape: (n_rel, emb_dim, emb_dim),
        dtype: torch.float
            Matrices of relations.

        """
        self.normalize_parameters()
        return self.ent_emb.weight.data, \
            self.rel_mat.weight.data.view(-1, self.emb_dim, self.emb_dim)

    def lp_scoring_function(self, h, t, r):
        """Link prediction evaluation helper function. See
        torchkge.models.interfaces.Models for more details of the API.

        """
        b_size = h.shape[0]

        if len(h.shape) == 2 and len(t.shape) == 3:
            # this is the tail completion case in link prediction
            h = h.view(b_size, 1, self.emb_dim)
            hr = matmul(h, r).view(b_size, self.emb_dim, 1)
            return (hr * t.transpose(1, 2)).sum(dim=1)
        else:
            # this is the head completion case in link prediction
            t = t.view(b_size, self.emb_dim, 1)
            return (h.transpose(1, 2) * matmul(r, t)).sum(dim=1)

    def lp_prep_cands(self, h_idx, t_idx, r_idx):
        """Link prediction evaluation helper function. Get entities embeddings
        and relations embeddings. The output will be fed to the
        `lp_scoring_function` method. See torchkge.models.interfaces.Models for
        more details on the API.

        """
        b_size = h_idx.shape[0]

        h_emb = self.ent_emb(h_idx)
        t_emb = self.ent_emb(t_idx)
        r_mat = self.rel_mat(r_idx).view(-1, self.emb_dim, self.emb_dim)

        candidates = self.ent_emb.weight.data.view(1, self.n_ent, self.emb_dim)
        candidates = candidates.expand(b_size, self.n_ent, self.emb_dim)

        return h_emb, t_emb, candidates, r_mat


class DistMultModel(BilinearModel):
    """Implementation of DistMult model detailed in 2014 paper by Yang et al..
    This class inherits from the
    :class:`torchkge.models.interfaces.BilinearModel` interface. It then has
    its attributes as well.

    References
    ----------
    * Bishan Yang, Wen-tau Yih, Xiaodong He, Jianfeng Gao, and Li Deng.
      `Embedding Entities and Relations for Learning and Inference in
      Knowledge Bases. <https://arxiv.org/abs/1412.6575>`_
      arXiv :1412.6575 [cs], December 2014.


    Parameters
    ----------
    emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    ent_emb: torch.nn.Embedding, shape: (n_ent, emb_dim)
        Embeddings of the entities, initialized with Xavier uniform
        distribution and then normalized.
    rel_emb: torch.nn.Embedding, shape: (n_rel, emb_dim)
        Embeddings of the relations, initialized with Xavier uniform
        distribution.

    """

    def __init__(self, emb_dim, n_entities, n_relations):
        super().__init__(emb_dim, n_entities, n_relations)
        self.emb_dim = emb_dim

        self.ent_emb = init_embedding(self.n_ent, self.emb_dim)
        self.rel_emb = init_embedding(self.n_rel, self.emb_dim)

        self.normalize_parameters()

    def scoring_function(self, h_idx, t_idx, r_idx):
        """Compute the scoring function for the triplets given as argument:
        :math:`h^T \\cdot diag(r) \\cdot t`. See referenced paper for more
        details on the score. See torchkge.models.interfaces.Models for more
        details on the API.

        """
        h = normalize(self.ent_emb(h_idx), p=2, dim=1)
        t = normalize(self.ent_emb(t_idx), p=2, dim=1)
        r = self.rel_emb(r_idx)

        return (h * r * t).sum(dim=1)

    def normalize_parameters(self):
        """Normalize the entity embeddings, as explained in original paper.
        This methods should be called at the end of each training epoch and at
        the end of training as well.

        """
        self.ent_emb.weight.data = normalize(self.ent_emb.weight.data, p=2,
                                             dim=1)

    def get_embeddings(self):
        """Return the embeddings of entities and relations.

        Returns
        -------
        ent_emb: torch.Tensor, shape: (n_ent, emb_dim), dtype: torch.float
            Embeddings of entities.
        rel_emb: torch.Tensor, shape: (n_rel, emb_dim), dtype: torch.float
            Embeddings of relations.

        """
        self.normalize_parameters()
        return self.ent_emb.weight.data, self.rel_emb.weight.data

    def lp_scoring_function(self, h, t, r):
        """Link prediction evaluation helper function. See
        torchkge.models.interfaces.Models for more details on the API.

        """
        b_size = h.shape[0]

        if len(h.shape) == 2 and len(t.shape) == 3:
            # this is the tail completion case in link prediction
            hr = (h * r).view(b_size, self.emb_dim, 1)
            return (hr * t.transpose(1, 2)).sum(dim=1)
        else:
            # this is the head completion case in link prediction
            rt = (r * t).view(b_size, self.emb_dim, 1)
            return (h.transpose(1, 2) * rt).sum(dim=1)

    def lp_prep_cands(self, h_idx, t_idx, r_idx):
        """Link prediction evaluation helper function. Get entities embeddings
        and relations embeddings. The output will be fed to the
        `lp_scoring_function` method. See torchkge.models.interfaces.Models for
        more details on the API.

        """
        b_size = h_idx.shape[0]

        h = self.ent_emb(h_idx)
        t = self.ent_emb(t_idx)
        r = self.rel_emb(r_idx)

        candidates = self.ent_emb.weight.data.view(1, self.n_ent, self.emb_dim)
        candidates = candidates.expand(b_size, self.n_ent, self.emb_dim)

        return h, t, candidates, r


class HolEModel(BilinearModel):
    """Implementation of HolE model detailed in 2015 paper by Nickel et al..
    This class inherits from the
    :class:`torchkge.models.interfaces.BilinearModel` interface. It then has
    its attributes as well.

    References
    ----------
    * Maximilian Nickel, Lorenzo Rosasco, and Tomaso Poggio.
      `Holographic Embeddings of Knowledge Graphs.
      <https://arxiv.org/abs/1510.04935>`_
      arXiv :1510.04935 [cs, stat], October 2015.

    Parameters
    ----------
    emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    ent_emb: torch.nn.Embedding, shape: (n_ent, emb_dim)
        Embeddings of the entities, initialized with Xavier uniform
        distribution and then normalized.
    rel_emb: torch.nn.Embedding, shape: (n_rel, emb_dim)
        Embeddings of the relations, initialized with Xavier uniform
        distribution.

    """

    def __init__(self, emb_dim, n_entities, n_relations):
        super().__init__(emb_dim, n_entities, n_relations)

        self.ent_emb = init_embedding(self.n_ent, self.emb_dim)
        self.rel_emb = init_embedding(self.n_rel, self.emb_dim)

        self.normalize_parameters()

    def scoring_function(self, h_idx, t_idx, r_idx):
        """Compute the scoring function for the triplets given as argument:
        :math:`h^T \\cdot M_r \\cdot t` where :math:`M_r` is the rolling matrix
        built from the relation embedding `r`. See referenced paper for more
        details on the score. See torchkge.models.interfaces.Models for more
        details on the API.

        """
        h = normalize(self.ent_emb(h_idx), p=2, dim=1)
        t = normalize(self.ent_emb(t_idx), p=2, dim=1)
        r = self.get_rolling_matrix(self.rel_emb(r_idx))
        hr = matmul(h.view(-1, 1, self.emb_dim), r)
        return (hr.view(-1, self.emb_dim) * t).sum(dim=1)

    @staticmethod
    def get_rolling_matrix(x):
        """Build a rolling matrix.

        Parameters
        ----------
        x: torch.Tensor, shape: (b_size, dim)

        Returns
        -------
        mat: torch.Tensor, shape: (b_size, dim, dim)
            Rolling matrix such that mat[i,j] = x[i - j mod(dim)]
        """
        b_size, dim = x.shape
        x = x.view(b_size, 1, dim)
        return cat([x.roll(i, dims=2) for i in range(dim)], dim=1)

    def normalize_parameters(self):
        """Normalize the entity embeddings, as explained in original paper.
        This methods should be called at the end of each training epoch and at
        the end of training as well.

        """
        self.ent_emb.weight.data = normalize(self.ent_emb.weight.data, p=2,
                                             dim=1)

    def get_embeddings(self):
        """Return the embeddings of entities and relations.

        Returns
        -------
        ent_emb: torch.Tensor, shape: (n_ent, emb_dim), dtype: torch.float
            Embeddings of entities.
        rel_emb: torch.Tensor, shape: (n_rel, emb_dim), dtype: torch.float
            Embeddings of relations.

        """
        self.normalize_parameters()
        return self.ent_emb.weight.data, self.rel_emb.weight.data

    def lp_scoring_function(self, h, t, r):
        """Link prediction evaluation helper function. See
        torchkge.models.interfaces.Models for more details on the API.

        """
        b_size = h.shape[0]

        if len(h.shape) == 2 and len(t.shape) == 3:
            # this is the tail completion case in link prediction
            h = h.view(b_size, 1, self.emb_dim)
            hr = matmul(h, r).view(b_size, self.emb_dim, 1)
            return (hr * t.transpose(1, 2)).sum(dim=1)
        else:
            # this is the head completion case in link prediction
            t = t.view(b_size, self.emb_dim, 1)
            return (h.transpose(1, 2) * matmul(r, t)).sum(dim=1)

    def lp_prep_cands(self, h_idx, t_idx, r_idx):
        """Link prediction evaluation helper function. Get entities embeddings
        and relations embeddings. The output will be fed to the
        `lp_scoring_function` method. See torchkge.models.interfaces.Models
        for more details on the API.

        """
        b_size = h_idx.shape[0]
        h_emb = self.ent_emb(h_idx)
        t_emb = self.ent_emb(t_idx)
        r_mat = self.get_rolling_matrix(self.rel_emb(r_idx))

        candidates = self.ent_emb.weight.data.view(1, self.n_ent, self.emb_dim)
        candidates = candidates.expand(b_size, self.n_ent, self.emb_dim)

        return h_emb, t_emb, candidates, r_mat


class ComplExModel(BilinearModel):
    """Implementation of ComplEx model detailed in 2016 paper by Trouillon et
    al.. This class inherits from the
    :class:`torchkge.models.interfaces.BilinearModel` interface. It then has
    its attributes as well.

    References
    ----------
    * Théo Trouillon, Johannes Welbl, Sebastian Riedel, Éric Gaussier, and
      Guillaume Bouchard.
      `Complex Embeddings for Simple Link Prediction.
      <https://arxiv.org/abs/1606.06357>`_
      arXiv :1606.06357 [cs, stat], June 2016.

    Parameters
    ----------
    emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    re_ent_emb: torch.nn.Embedding, shape: (n_ent, emb_dim)
        Real part of the entities complex embeddings. Initialized with Xavier
        uniform distribution.
    im_ent_emb: torch.nn.Embedding, shape: (n_ent, emb_dim)
        Imaginary part of the entities complex embeddings. Initialized with
        Xavier uniform distribution.
    re_rel_emb: torch.nn.Embedding, shape: (n_rel, emb_dim)
        Real part of the relations complex embeddings. Initialized with Xavier
        uniform distribution.
    im_rel_emb: torch.nn.Embedding, shape: (n_rel, emb_dim)
        Imaginary part of the relations complex embeddings. Initialized with
        Xavier uniform distribution.
    """

    def __init__(self, emb_dim, n_entities, n_relations):
        super().__init__(emb_dim, n_entities, n_relations)
        self.re_ent_emb = init_embedding(self.n_ent, self.emb_dim)
        self.im_ent_emb = init_embedding(self.n_ent, self.emb_dim)
        self.re_rel_emb = init_embedding(self.n_rel, self.emb_dim)
        self.im_rel_emb = init_embedding(self.n_rel, self.emb_dim)

    def scoring_function(self, h_idx, t_idx, r_idx):
        """Compute the real part of the Hermitian product
        :math:`\\Re(h^T \\cdot diag(r) \\cdot \\bar{t})` for each sample of
        the batch. See referenced paper for more details on the score. See
        torchkge.models.interfaces.Models for more details on the API.

        """

        re_h, im_h = self.re_ent_emb(h_idx), self.im_ent_emb(h_idx)
        re_t, im_t = self.re_ent_emb(t_idx), self.im_ent_emb(t_idx)
        re_r, im_r = self.re_rel_emb(r_idx), self.im_rel_emb(r_idx)

        return (re_h * (re_r * re_t + im_r * im_t) + im_h * (
                    re_r * im_t - im_r * re_t)).sum(dim=1)

    def normalize_parameters(self):
        """According to original paper, the embeddings should not be
        normalized.

        """
        pass

    def get_embeddings(self):
        """Return the embeddings of entities and relations.

        Returns
        -------
        re_ent_emb: torch.Tensor, shape: (n_ent, emb_dim), dtype: torch.float
            Real part of embeddings of entities.
        im_ent_emb: torch.Tensor, shape: (n_ent, emb_dim), dtype: torch.float
            Imaginary part of embeddings of entities.
        re_rel_emb: torch.Tensor, shape: (n_rel, emb_dim), dtype: torch.float
            Real part of embeddings of relations.
        im_rel_emb: torch.Tensor, shape: (n_rel, emb_dim), dtype: torch.float
            Imaginary part of embeddings of relations.

        """
        self.normalize_parameters()
        return self.re_ent_emb.weight.data, self.im_ent_emb.weight.data,\
            self.re_rel_emb.weight.data, self.im_rel_emb.weight.data

    def lp_scoring_function(self, h, t, r):
        """Link prediction evaluation helper function. See
        torchkge.models.interfaces.Models for more details one the API.

        """
        re_h, im_h = h[0], h[1]
        re_t, im_t = t[0], t[1]
        re_r, im_r = r[0], r[1]
        b_size = re_h.shape[0]

        if len(re_h.shape) == 2 and len(re_t.shape) == 3:
            # this is the tail completion case in link prediction
            return ((re_h * re_r).view(b_size, self.emb_dim,
                                       1) * re_t.transpose(1, 2)
                    + (re_h * im_r).view(b_size, self.emb_dim,
                                         1) * im_t.transpose(1, 2)
                    + (im_h * re_r).view(b_size, self.emb_dim,
                                         1) * im_t.transpose(1, 2)
                    - (im_h * im_r).view(b_size, self.emb_dim,
                                         1) * re_t.transpose(1, 2)).sum(dim=1)

        if len(re_h.shape) == 3 and len(re_t.shape) == 2:
            # this is the head completion case in link prediction
            return (re_h.transpose(1, 2) * (re_r * re_t + im_r * im_t).view(
                b_size, self.emb_dim, 1)
                    + im_h.transpose(1, 2) * (re_r * im_t - im_r * re_t).view(
                        b_size, self.emb_dim, 1)).sum(dim=1)

    def lp_prep_cands(self, h_idx, t_idx, r_idx):
        """Link prediction evaluation helper function. Get entities embeddings
        and relations embeddings. The output will be fed to the
        `lp_scoring_function` method. See torchkge.models.interfaces.Models for
        more details on the API.

        """
        b_size = h_idx.shape[0]

        re_candidates = self.re_ent_emb.weight.data.view(1, self.n_ent,
                                                         self.emb_dim)
        re_candidates = re_candidates.expand(b_size, self.n_ent, self.emb_dim)

        im_candidates = self.im_ent_emb.weight.data.view(1, self.n_ent,
                                                         self.emb_dim)
        im_candidates = im_candidates.expand(b_size, self.n_ent, self.emb_dim)

        re_h, im_h = self.re_ent_emb(h_idx), self.im_ent_emb(h_idx)
        re_t, im_t = self.re_ent_emb(t_idx), self.im_ent_emb(t_idx)
        re_r, im_r = self.re_rel_emb(r_idx), self.im_rel_emb(r_idx)

        return (re_h, im_h), (re_t, im_t), (re_candidates, im_candidates), (
            re_r, im_r)


class AnalogyModel(BilinearModel):
    """Implementation of ANALOGY model detailed in 2017 paper by Liu et al..
    According to their remark in the implementation details, the number of
    scalars on the diagonal of each relation-specific matrix is by default set
    to be half the embedding dimension. This class inherits from the
    :class:`torchkge.models.interfaces.BilinearModel` interface. It then has
    its attributes as well.

    References
    ----------
    * Hanxiao Liu, Yuexin Wu, and Yiming Yang.
      `Analogical Inference for Multi-Relational Embeddings.
      <https://arxiv.org/abs/1705.02426>`_
      arXiv :1705.02426 [cs], May 2017.

    Parameters
    ----------
    emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.
    scalar_share: float
        Share of the diagonal elements of the relation-specific matrices to be
        scalars. By default it is set to half according to the original paper.

    Attributes
    ----------
    scalar_dim: int
        Number of diagonal elements of the relation-specific matrices to be
        scalars. By default it is set to half the embedding dimension according
        to the original paper.
    complex_dim: int
        Number of 2x2 matrices on the diagonals of relation-specific matrices.
    sc_ent_emb: torch.nn.Embedding, shape: (n_ent, emb_dim)
        Part of the entities embeddings associated to the scalar part of the
        relation specific matrices. Initialized with Xavier uniform
        distribution.
    re_ent_emb: torch.nn.Embedding, shape: (n_ent, emb_dim)
        Real part of the entities complex embeddings. Initialized with Xavier
        uniform distribution. As explained in the authors' paper, almost
        diagonal matrices can be seen as complex matrices.
    im_ent_emb: torch.nn.Embedding, shape: (n_ent, emb_dim)
        Imaginary part of the entities complex embeddings. Initialized with
        Xavier uniform distribution. As explained in the authors' paper, almost
        diagonal matrices can be seen as complex matrices.
    sc_rel_emb: torch.nn.Embedding, shape: (n_rel, emb_dim)
        Part of the entities embeddings associated to the scalar part of the
        relation specific matrices. Initialized with Xavier uniform
        distribution.
    re_rel_emb: torch.nn.Embedding, shape: (n_rel, emb_dim)
        Real part of the relations complex embeddings. Initialized with Xavier
        uniform distribution. As explained in the authors' paper, almost
        diagonal matrices can be seen as complex matrices.
    im_rel_emb: torch.nn.Embedding, shape: (n_rel, emb_dim)
        Imaginary part of the relations complex embeddings. Initialized with
        Xavier uniform distribution. As explained in the authors' paper, almost
        diagonal matrices can be seen as complex matrices.
    """

    def __init__(self, emb_dim, n_entities, n_relations, scalar_share=0.5):
        super().__init__(emb_dim, n_entities, n_relations)

        self.scalar_dim = int(self.emb_dim * scalar_share)
        self.complex_dim = int((self.emb_dim - self.scalar_dim))

        self.sc_ent_emb = init_embedding(self.n_ent, self.scalar_dim)
        self.re_ent_emb = init_embedding(self.n_ent, self.complex_dim)
        self.im_ent_emb = init_embedding(self.n_ent, self.complex_dim)

        self.sc_rel_emb = init_embedding(self.n_rel, self.scalar_dim)
        self.re_rel_emb = init_embedding(self.n_rel, self.complex_dim)
        self.im_rel_emb = init_embedding(self.n_rel, self.complex_dim)

    def scoring_function(self, h_idx, t_idx, r_idx):
        """Compute the scoring function for the triplets given as argument:
        :math:`h_{sc}^T \\cdot diag(r_{sc}) \\cdot t_{sc} + \\Re(h_{compl}
        \\cdot diag(r_{compl} \\cdot t_{compl}))`. See referenced paper for
        more details on the score. See torchkge.models.interfaces.Models for
        more details on the API.
        """

        sc_h, re_h, im_h = self.sc_ent_emb(h_idx), self.re_ent_emb(
            h_idx), self.im_ent_emb(h_idx)
        sc_t, re_t, im_t = self.sc_ent_emb(t_idx), self.re_ent_emb(
            t_idx), self.im_ent_emb(t_idx)
        sc_r, re_r, im_r = self.sc_rel_emb(r_idx), self.re_rel_emb(
            r_idx), self.im_rel_emb(r_idx)

        return ((sc_h * sc_r * sc_t).sum(dim=1) +
                (re_h * (re_r * re_t + im_r * im_t) + im_h * (
                            re_r * im_t - im_r * re_t)).sum(dim=1))

    def normalize_parameters(self):
        """According to original paper, the embeddings should not be
        normalized.
        """
        pass

    def get_embeddings(self):
        """Return the embeddings of entities and relations.

        Returns
        -------
        sc_ent_emb: torch.Tensor, shape: (n_ent, emb_dim), dtype: torch.float
            Scalar part of embeddings of entities.
        re_ent_emb: torch.Tensor, shape: (n_ent, emb_dim), dtype: torch.float
            Real part of embeddings of entities.
        im_ent_emb: torch.Tensor, shape: (n_ent, emb_dim), dtype: torch.float
            Imaginary part of embeddings of entities.
        sc_rel_emb: torch.Tensor, shape: (n_rel, emb_dim), dtype: torch.float
            Scalar part of embeddings of relations.
        re_rel_emb: torch.Tensor, shape: (n_rel, emb_dim), dtype: torch.float
            Real part of embeddings of relations.
        im_rel_emb: torch.Tensor, shape: (n_rel, emb_dim), dtype: torch.float
            Imaginary part of embeddings of relations.

        """
        self.normalize_parameters()
        return self.sc_ent_emb.weight.data, self.re_ent_emb.weight.data, \
            self.im_ent_emb.weight.data, self.sc_rel_emb.weight.data, \
            self.re_rel_emb.weight.data, self.im_rel_emb.weight.data

    def lp_scoring_function(self, h, t, r):
        """Link prediction evaluation helper function. See
        torchkge.models.interfaces.Models for more details one the API.

        """
        sc_h, re_h, im_h = h[0], h[1], h[2]
        sc_t, re_t, im_t = t[0], t[1], t[2]
        sc_r, re_r, im_r = r[0], r[1], r[2]
        b_size = re_h.shape[0]

        if len(re_h.shape) == 2 and len(re_t.shape) == 3:
            # this is the tail completion case in link prediction
            return (((sc_h * sc_r).view(b_size,
                                        self.scalar_dim,
                                        1) * sc_t.transpose(1, 2)).sum(dim=1)
                    + ((re_h * re_r).view(b_size,
                                          self.complex_dim,
                                          1) * re_t.transpose(1, 2)).sum(dim=1)
                    + ((re_h * im_r).view(b_size,
                                          self.complex_dim,
                                          1) * im_t.transpose(1, 2)).sum(dim=1)
                    + ((im_h * re_r).view(b_size,
                                          self.complex_dim,
                                          1) * im_t.transpose(1, 2)).sum(dim=1)
                    - ((im_h * im_r).view(b_size,
                                          self.complex_dim,
                                          1) * re_t.transpose(1, 2)).sum(dim=1)
                    )

        if len(re_h.shape) == 3 and len(re_t.shape) == 2:
            # this is the head completion case in link prediction
            sc_rt = (sc_r * sc_t).view(b_size, self.scalar_dim, 1)
            sum1 = re_h.transpose(1, 2) * (re_r * re_t + im_r * im_t).view(
                        b_size, self.complex_dim, 1)
            sum2 = im_h.transpose(1, 2) * (re_r * im_t - im_r * re_t).view(
                        b_size, self.complex_dim, 1)
            return ((sc_h.transpose(1, 2) * sc_rt).sum(dim=1)
                    + sum1.sum(dim=1)
                    + sum2.sum(dim=1))

    def lp_prep_cands(self, h_idx, t_idx, r_idx):
        """Link prediction evaluation helper function. Get entities embeddings
        and relations embeddings. The output will be fed to the
        `lp_scoring_function` method. See torchkge.models.interfaces.Models for
        more details on the API.

        """
        b_size = h_idx.shape[0]

        sc_candidates = self.sc_ent_emb.weight.data
        sc_candidates = sc_candidates.view(1, self.n_ent, self.scalar_dim)
        sc_candidates = sc_candidates.expand(b_size,
                                             self.n_ent,
                                             self.scalar_dim)

        re_candidates = self.re_ent_emb.weight.data
        re_candidates = re_candidates.view(1, self.n_ent, self.complex_dim)
        re_candidates = re_candidates.expand(b_size,
                                             self.n_ent,
                                             self.complex_dim)

        im_candidates = self.im_ent_emb.weight.data
        im_candidates = im_candidates.view(1, self.n_ent, self.complex_dim)
        im_candidates = im_candidates.expand(b_size,
                                             self.n_ent,
                                             self.complex_dim)

        sc_h = self.sc_ent_emb(h_idx)
        re_h = self.re_ent_emb(h_idx)
        im_h = self.im_ent_emb(h_idx)

        sc_t = self.sc_ent_emb(t_idx)
        re_t = self.re_ent_emb(t_idx)
        im_t = self.im_ent_emb(t_idx)

        sc_r = self.sc_rel_emb(r_idx)
        re_r = self.re_rel_emb(r_idx)
        im_r = self.im_rel_emb(r_idx)

        return (sc_h, re_h, im_h), \
               (sc_t, re_t, im_t), \
               (sc_candidates, re_candidates, im_candidates), \
               (sc_r, re_r, im_r)



from .data import DataLoader, get_data_home, clear_data_home

from .data_redundancy import count_triplets, duplicates
from .data_redundancy import cartesian_product_relations

from .datasets import load_fb15k, load_fb13, load_fb15k237
from .datasets import load_wn18, load_wn18rr
from .datasets import load_yago3_10, load_wikidatasets, load_wikidata_vitals

from .dissimilarities import l1_dissimilarity, l2_dissimilarity
from .dissimilarities import l1_torus_dissimilarity, l2_torus_dissimilarity, \
    el2_torus_dissimilarity

from .losses import MarginLoss, LogisticLoss, BinaryCrossEntropyLoss
from .modeling import init_embedding, get_true_targets, load_embeddings
from .operations import get_rank, get_mask, get_bernoulli_probs
from .training import Trainer, TrainDataLoader



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from pandas import DataFrame

from torch import zeros, cat


def get_mask(length, start, end):
    """Create a mask of length `length` filled with 0s except between indices
    `start` (included) and `end` (excluded).

    Parameters
    ----------
    length: int
        Length of the mask to be created.
    start: int
        First index (included) where the mask will be filled with 0s.
    end: int
        Last index (excluded) where the mask will be filled with 0s.

    Returns
    -------
    mask: `torch.Tensor`, shape: (length), dtype: `torch.bool`
        Mask of length `length` filled with 0s except between indices `start`
        (included) and `end` (excluded).
    """
    mask = zeros(length)
    mask[[i for i in range(start, end)]] = 1
    return mask.bool()


def get_rank(data, true, low_values=False):
    """Computes the rank of entity at index true[i]. If the rank is k then
    there are k-1 entities with better (higher or lower) value in data.

    Parameters
    ----------
    data: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts, dimensions)
        Scores for each entity.
    true: `torch.Tensor`, dtype: `torch.int`, shape: (n_facts)
        true[i] is the index of the true entity for test i of the batch.
    low_values: bool, optional (default=False)
        if True, best rank is the lowest score else it is the highest.

    Returns
    -------
    ranks: `torch.Tensor`, dtype: `torch.int`, shape: (n_facts)
        ranks[i] - 1 is the number of entities which have better (or same)
        scores in data than the one and index true[i]
    """
    true_data = data.gather(1, true.long().view(-1, 1))

    if low_values:
        return (data <= true_data).sum(dim=1)
    else:
        return (data >= true_data).sum(dim=1)


def get_dictionaries(df, ent=True):
    """Build entities or relations dictionaries.

    Parameters
    ----------
    df: `pandas.DataFrame`
        Data frame containing three columns [from, to, rel].
    ent: bool
        if True then ent2ix is returned, if False then rel2ix is returned.

    Returns
    -------
    dict: dictionary
        Either ent2ix or rel2ix.

    """
    if ent:
        tmp = list(set(df['from'].unique()).union(set(df['to'].unique())))
        return {ent: i for i, ent in enumerate(sorted(tmp))}
    else:
        tmp = list(df['rel'].unique())
        return {rel: i for i, rel in enumerate(sorted(tmp))}


def get_tph(t):
    """Get the average number of tail per heads for each relation.

    Parameters
    ----------
    t: `torch.Tensor`, dtype: `torch.long`, shape: (b_size, 3)
        First column contains head indices, second tails and third relations.
    Returns
    -------
    d: dict
        keys: relation indices, values: average number of tail per heads.
    """
    df = DataFrame(t.numpy(), columns=['from', 'to', 'rel'])
    df = df.groupby(['from', 'rel']).count().groupby('rel').mean()
    df.reset_index(inplace=True)
    return {df.loc[i].values[0]: df.loc[i].values[1] for i in df.index}


def get_hpt(t):
    """Get the average number of head per tails for each relation.

    Parameters
    ----------
    t: `torch.Tensor`, dtype: `torch.long`, shape: (b_size, 3)
        First column contains head indices, second tails and third relations.
    Returns
    -------
    d: dict
        keys: relation indices, values: average number of head per tails.
    """
    df = DataFrame(t.numpy(), columns=['from', 'to', 'rel'])
    df = df.groupby(['rel', 'to']).count().groupby('rel').mean()
    df.reset_index(inplace=True)
    return {df.loc[i].values[0]: df.loc[i].values[1] for i in df.index}


def get_bernoulli_probs(kg):
    """Evaluate the Bernoulli probabilities for negative sampling as in the
    TransH original paper by Wang et al. (2014).

    Parameters
    ----------
    kg: `torchkge.data_structures.KnowledgeGraph`

    Returns
    -------
    tph: dict
        keys: relations , values: sampling probabilities as described by
        Wang et al. in their paper.

    """
    t = cat((kg.head_idx.view(-1, 1),
             kg.tail_idx.view(-1, 1),
             kg.relations.view(-1, 1)), dim=1)

    hpt = get_hpt(t)
    tph = get_tph(t)

    assert hpt.keys() == tph.keys()

    for k in tph.keys():
        tph[k] = tph[k] / (tph[k] + hpt[k])

    return tph



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from ..exceptions import NoPreTrainedVersionError
from ..models import TransEModel, ComplExModel
from ..utils import load_embeddings


def load_pretrained_transe(dataset, emb_dim, data_home=None):
    """Load a pretrained version of TransE model.

    Parameters
    ----------
    dataset: str
    emb_dim: int
        Embedding dimension
    data_home: str (opt, default None)
        Path to the `torchkge_data` directory (containing data folders). Useful
        for pre-trained model loading.

    Returns
    -------
    model: `TorchKGE.model.translation.TransEModel`
        Pretrained version of TransE model.
    """
    try:
        assert (dataset in {'fb15k', 'fb15k237', 'wn18rr'} and emb_dim == 100) \
            or (dataset == 'fb15k237' and emb_dim == 150) \
            or (dataset == 'fb15k237' and emb_dim == 200)
    except AssertionError:
        raise NoPreTrainedVersionError('No pre-trained version of TransE for '
                                       '{} in dimension {}'.format(dataset,
                                                                   emb_dim))

    state_dict = load_embeddings('transe', emb_dim, dataset, data_home)
    model = TransEModel(emb_dim,
                        n_entities=state_dict['ent_emb.weight'].shape[0],
                        n_relations=state_dict['rel_emb.weight'].shape[0],
                        dissimilarity_type='L2')
    model.load_state_dict(state_dict)

    return model


def load_pretrained_complex(dataset, emb_dim, data_home=None):
    """Load a pretrained version of ComplEx model.

    Parameters
    ----------
    dataset: str
    emb_dim: int
        Embedding dimension
    data_home: str (opt, default None)
        Path to the `torchkge_data` directory (containing data folders). Useful
        for pre-trained model loading.

    Returns
    -------
    model: `TorchKGE.model.translation.ComplExModel`
        Pretrained version of ComplEx model.
    """
    try:
        assert (dataset == 'wn18rr' and emb_dim == 200)
    except AssertionError:
        raise NoPreTrainedVersionError('No pre-trained version of TransE for '
                                       '{} in dimension {}'.format(dataset,
                                                                   emb_dim))

    state_dict = load_embeddings('complex', emb_dim, dataset, data_home)
    model = ComplExModel(emb_dim,
                         n_entities=state_dict['re_ent_emb.weight'].shape[0],
                         n_relations=state_dict['re_rel_emb.weight'].shape[0])
    model.load_state_dict(state_dict)

    return model



import pandas as pd
import unittest

from torch import Tensor, int64

from torchkge.data_structures import KnowledgeGraph
from torchkge.exceptions import WrongArgumentsError, SanityError, SizeMismatchError


class TestUtils(unittest.TestCase):
    """Tests for `torchkge.utils`."""

    def setUp(self):
        self.df = pd.DataFrame([[0, 1, 0], [0, 2, 0], [0, 3, 0], [0, 4, 0], [1, 2, 1], [1, 3, 2], [2, 4, 0], [3, 4, 4],
                                [5, 4, 0]], columns=['from', 'to', 'rel'])
        self.kg = KnowledgeGraph(self.df)

    def test_KnowledgeGraph_Builder(self):
        assert len(self.kg) == 9
        assert self.kg.n_ent == 6
        assert self.kg.n_rel == 4
        assert (type(self.kg.rel2ix) == dict) & (type(self.kg.rel2ix) == dict)
        assert (type(self.kg.head_idx) == Tensor) & (type(self.kg.tail_idx) == Tensor) & \
               (type(self.kg.relations) == Tensor)
        assert (self.kg.head_idx.dtype == int64) & (self.kg.tail_idx.dtype == int64) & \
               (self.kg.relations.dtype == int64)
        assert (len(self.kg.head_idx) == len(self.kg.tail_idx) == len(self.kg.relations))

        kg_dict = {'heads': self.kg.head_idx, 'tails': self.kg.tail_idx, 'relations': self.kg.relations}
        with self.assertRaises(WrongArgumentsError):
            KnowledgeGraph()
        with self.assertRaises(WrongArgumentsError):
            KnowledgeGraph(kg=kg_dict, df=self.df)
        with self.assertRaises(WrongArgumentsError):
            KnowledgeGraph(kg=kg_dict)
        with self.assertRaises(WrongArgumentsError):
            KnowledgeGraph(kg={'heads': self.kg.head_idx, 'tails': self.kg.tail_idx},
                           ent2ix=self.kg.ent2ix, rel2ix=self.kg.rel2ix)
        with self.assertRaises(SanityError):
            KnowledgeGraph(kg={'heads': self.kg.head_idx[:-1],
                               'tails': self.kg.tail_idx,
                               'relations': self.kg.relations},
                           ent2ix=self.kg.ent2ix, rel2ix=self.kg.rel2ix)
        with self.assertRaises(SanityError):
            KnowledgeGraph(kg={'heads': self.kg.head_idx.int(),
                               'tails': self.kg.tail_idx,
                               'relations': self.kg.relations},
                           ent2ix=self.kg.ent2ix, rel2ix=self.kg.rel2ix)

    def test_split_kg(self):
        assert (len(self.kg.split_kg()) == 2) & (len(self.kg.split_kg(validation=True)) == 3)
        with self.assertRaises(SizeMismatchError):
            self.kg.split_kg(sizes=(1, 2, 3, 4))
        with self.assertRaises(WrongArgumentsError):
            self.kg.split_kg(sizes=(9, 9, 9))
        with self.assertRaises(WrongArgumentsError):
            self.kg.split_kg(sizes=(9, 9))




# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from collections import defaultdict

from torch import tensor, bernoulli, randint, ones, rand, cat

from torchkge.exceptions import NotYetImplementedError
from torchkge.utils.data import DataLoader
from torchkge.utils.operations import get_bernoulli_probs


class NegativeSampler:
    """This is an interface for negative samplers in general.

    Parameters
    ----------
    kg: torchkge.data_structures.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data_structures.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data_structures.KnowledgeGraph (optional)
        Test knowledge graph.
    n_neg: int
        Number of negative sample to create from each fact.

    Attributes
    ----------
    kg: torchkge.data_structures.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data_structures.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data_structures.KnowledgeGraph (optional)
        Test knowledge graph.
    n_ent: int
        Number of entities in the entire knowledge graph. This is the same in
        `kg`, `kg_val` and `kg_test`.
    n_facts: int
        Number of triplets in `kg`.
    n_facts_val: in
        Number of triplets in `kg_val`.
    n_facts_test: int
        Number of triples in `kg_test`.
    n_neg: int
        Number of negative sample to create from each fact.
    """

    def __init__(self, kg, kg_val=None, kg_test=None, n_neg=1):
        self.kg = kg
        self.n_ent = kg.n_ent
        self.n_facts = kg.n_facts

        self.kg_val = kg_val
        self.kg_test = kg_test

        self.n_neg = n_neg

        if kg_val is None:
            self.n_facts_val = 0
        else:
            self.n_facts_val = kg_val.n_facts

        if kg_test is None:
            self.n_facts_test = 0
        else:
            self.n_facts_test = kg_test.n_facts

    def corrupt_batch(self, heads, tails, relations, n_neg):
        raise NotYetImplementedError('NegativeSampler is just an interface, '
                                     'please consider using a child class '
                                     'where this is implemented.')

    def corrupt_kg(self, batch_size, use_cuda, which='main'):
        """Corrupt an entire knowledge graph using a dataloader and by calling
        `corrupt_batch` method.

        Parameters
        ----------
        batch_size: int
            Size of the batches used in the dataloader.
        use_cuda: bool
            Indicate whether to use cuda or not
        which: str
            Indicate which graph should be corrupted. Possible values are :
            * 'main': attribute self.kg is corrupted
            * 'train': attribute self.kg is corrupted
            * 'val': attribute self.kg_val is corrupted. In this case this
            attribute should have been initialized.
            * 'test': attribute self.kg_test is corrupted. In this case this
            attribute should have been initialized.

        Returns
        -------
        neg_heads: torch.Tensor, dtype: torch.long, shape: (n_facts)
            Tensor containing the integer key of negatively sampled heads of
            the relations in the graph designated by `which`.
        neg_tails: torch.Tensor, dtype: torch.long, shape: (n_facts)
            Tensor containing the integer key of negatively sampled tails of
            the relations in the graph designated by `which`.
        """
        assert which in ['main', 'train', 'test', 'val']
        if which == 'val':
            assert self.n_facts_val > 0
        if which == 'test':
            assert self.n_facts_test > 0

        if use_cuda:
            tmp_cuda = 'batch'
        else:
            tmp_cuda = None

        if which == 'val':
            dataloader = DataLoader(self.kg_val, batch_size=batch_size,
                                    use_cuda=tmp_cuda)
        elif which == 'test':
            dataloader = DataLoader(self.kg_test, batch_size=batch_size,
                                    use_cuda=tmp_cuda)
        else:
            dataloader = DataLoader(self.kg, batch_size=batch_size,
                                    use_cuda=tmp_cuda)

        corr_heads, corr_tails = [], []

        for i, batch in enumerate(dataloader):
            heads, tails, rels = batch[0], batch[1], batch[2]
            neg_heads, neg_tails = self.corrupt_batch(heads, tails, rels,
                                                      n_neg=1)

            corr_heads.append(neg_heads)
            corr_tails.append(neg_tails)

        if use_cuda:
            return cat(corr_heads).long().cpu(), cat(corr_tails).long().cpu()
        else:
            return cat(corr_heads).long(), cat(corr_tails).long()


class UniformNegativeSampler(NegativeSampler):
    """Uniform negative sampler as presented in 2013 paper by Bordes et al..
    Either the head or the tail of a triplet is replaced by another entity at
    random. The choice of head/tail is uniform. This class inherits from the
    :class:`torchkge.sampling.NegativeSampler` interface. It
    then has its attributes as well.

    References
    ----------
    * Antoine Bordes, Nicolas Usunier, Alberto Garcia-Duran, Jason Weston,
      and Oksana Yakhnenko.
      Translating Embeddings for Modeling Multi-relational Data.
      In Advances in Neural Information Processing Systems 26, pages 2787–2795,
      2013.
      https://papers.nips.cc/paper/5071-translating-embeddings-for-modeling-multi-relational-data

    Parameters
    ----------
    kg: torchkge.data_structures.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data_structures.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data_structures.KnowledgeGraph (optional)
        Test knowledge graph.
    n_neg: int
        Number of negative sample to create from each fact.
    """

    def __init__(self, kg, kg_val=None, kg_test=None, n_neg=1):
        super().__init__(kg, kg_val, kg_test, n_neg)

    def corrupt_batch(self, heads, tails, relations=None, n_neg=None):
        """For each true triplet, produce a corrupted one not different from
        any other true triplet. If `heads` and `tails` are cuda objects ,
        then the returned tensors are on the GPU.

        Parameters
        ----------
        heads: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of heads of the relations in the
            current batch.
        tails: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of tails of the relations in the
            current batch.
        relations: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of relations in the current
            batch. This is optional here and mainly present because of the
            interface with other NegativeSampler objects.
        n_neg: int (opt)
            Number of negative sample to create from each fact. It overwrites
            the value set at the construction of the sampler.
        Returns
        -------
        neg_heads: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of negatively sampled heads of
            the relations in the current batch.
        neg_tails: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of negatively sampled tails of
            the relations in the current batch.
        """
        if n_neg is None:
            n_neg = self.n_neg

        device = heads.device
        assert (device == tails.device)

        batch_size = heads.shape[0]
        neg_heads = heads.repeat(n_neg)
        neg_tails = tails.repeat(n_neg)

        # Randomly choose which samples will have head/tail corrupted
        mask = bernoulli(ones(size=(batch_size * n_neg,),
                              device=device) / 2).double()

        n_h_cor = int(mask.sum().item())
        neg_heads[mask == 1] = randint(1, self.n_ent,
                                       (n_h_cor,),
                                       device=device)
        neg_tails[mask == 0] = randint(1, self.n_ent,
                                       (batch_size * n_neg - n_h_cor,),
                                       device=device)

        return neg_heads.long(), neg_tails.long()


class BernoulliNegativeSampler(NegativeSampler):
    """Bernoulli negative sampler as presented in 2014 paper by Wang et al..
    Either the head or the tail of a triplet is replaced by another entity at
    random. The choice of head/tail is done using probabilities taking into
    account profiles of the relations. See the paper for more details. This
    class inherits from the
    :class:`torchkge.sampling.NegativeSampler` interface.
    It then has its attributes as well.

    References
    ----------
    * Zhen Wang, Jianwen Zhang, Jianlin Feng, and Zheng Chen.
      Knowledge Graph Embedding by Translating on Hyperplanes.
      In Twenty-Eighth AAAI Conference on Artificial Intelligence, June 2014.
      https://www.aaai.org/ocs/index.php/AAAI/AAAI14/paper/view/8531

    Parameters
    ----------
    kg: torchkge.data_structures.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data_structures.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data_structures.KnowledgeGraph (optional)
        Test knowledge graph.
    n_neg: int
        Number of negative sample to create from each fact.
    Attributes
    ----------
    bern_probs: torch.Tensor, dtype: torch.float, shape: (kg.n_rel)
        Bernoulli sampling probabilities. See paper for more details.

    """

    def __init__(self, kg, kg_val=None, kg_test=None, n_neg=1):
        super().__init__(kg, kg_val, kg_test, n_neg)
        self.bern_probs = self.evaluate_probabilities()

    def evaluate_probabilities(self):
        """Evaluate the Bernoulli probabilities for negative sampling as in the
        TransH original paper by Wang et al. (2014).
        """
        bern_probs = get_bernoulli_probs(self.kg)

        tmp = []
        for i in range(self.kg.n_rel):
            if i in bern_probs.keys():
                tmp.append(bern_probs[i])
            else:
                tmp.append(0.5)

        return tensor(tmp).float()

    def corrupt_batch(self, heads, tails, relations, n_neg=None):
        """For each true triplet, produce a corrupted one different from any
        other true triplet. If `heads` and `tails` are cuda objects , then the
        returned tensors are on the GPU.

        Parameters
        ----------
        heads: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of heads of the relations in the
            current batch.
        tails: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of tails of the relations in the
            current batch.
        relations: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of relations in the current
            batch.
        n_neg: int (opt)
            Number of negative sample to create from each fact. It overwrites
            the value set at the construction of the sampler.
        Returns
        -------
        neg_heads: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of negatively sampled heads of
            the relations in the current batch.
        neg_tails: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of negatively sampled tails of
            the relations in the current batch.
        """
        if n_neg is None:
            n_neg = self.n_neg

        device = heads.device
        assert (device == tails.device)

        batch_size = heads.shape[0]
        neg_heads = heads.repeat(n_neg)
        neg_tails = tails.repeat(n_neg)

        # Randomly choose which samples will have head/tail corrupted
        mask = bernoulli(self.bern_probs[relations].repeat(n_neg)).double()
        n_h_cor = int(mask.sum().item())
        neg_heads[mask == 1] = randint(1, self.n_ent,
                                       (n_h_cor,),
                                       device=device)
        neg_tails[mask == 0] = randint(1, self.n_ent,
                                       (batch_size * n_neg - n_h_cor,),
                                       device=device)

        return neg_heads.long(), neg_tails.long()


class PositionalNegativeSampler(BernoulliNegativeSampler):
    """Positional negative sampler as presented in 2011 paper by Socher et al..
    Either the head or the tail of a triplet is replaced by another entity
    chosen among entities that have already appeared at the same place in a
    triplet (involving the same relation). It is not clear in the paper how the
    choice of head/tail is done. We chose to use Bernoulli sampling as in 2014
    paper by Wang et al. as we believe it serves the same purpose as the
    original paper. This class inherits from the
    :class:`torchkge.sampling.BernouilliNegativeSampler` class
    seen as an interface. It then has its attributes as well.

    References
    ----------
    * Richard Socher, Danqi Chen, Christopher D Manning, and Andrew Ng.
      Reasoning With Neural Tensor Networks for Knowledge Base Completion.
      In Advances in Neural Information Processing Systems 26, pages 926–934.,
      2013.
      https://nlp.stanford.edu/pubs/SocherChenManningNg_NIPS2013.pdf
    * Zhen Wang, Jianwen Zhang, Jianlin Feng, and Zheng Chen.
      Knowledge Graph Embedding by Translating on Hyperplanes.
      In Twenty-Eighth AAAI Conference on Artificial Intelligence, June 2014.
      https://www.aaai.org/ocs/index.php/AAAI/AAAI14/paper/view/8531

    Parameters
    ----------
    kg: torchkge.data_structures.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data_structures.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data_structures.KnowledgeGraph (optional)
        Test knowledge graph.

    Attributes
    ----------
    possible_heads: dict
        keys : relations, values : list of possible heads for each relation.
    possible_tails: dict
        keys : relations, values : list of possible tails for each relation.
    n_poss_heads: list
        List of number of possible heads for each relation.
    n_poss_tails: list
        List of number of possible tails for each relation.

    """

    def __init__(self, kg, kg_val=None, kg_test=None):
        super().__init__(kg, kg_val, kg_test, 1)
        self.possible_heads, self.possible_tails, \
        self.n_poss_heads, self.n_poss_tails = self.find_possibilities()

    def find_possibilities(self):
        """For each relation of the knowledge graph (and possibly the
        validation graph but not the test graph) find all the possible heads
        and tails in the sense of Wang et al., e.g. all entities that occupy
        once this position in another triplet.

        Returns
        -------
        possible_heads: dict
            keys : relation index, values : list of possible heads
        possible tails: dict
            keys : relation index, values : list of possible tails
        n_poss_heads: torch.Tensor, dtype: torch.long, shape: (n_relations)
            Number of possible heads for each relation.
        n_poss_tails: torch.Tensor, dtype: torch.long, shape: (n_relations)
            Number of possible tails for each relation.

        """
        possible_heads, possible_tails = get_possible_heads_tails(self.kg)

        if self.n_facts_val > 0:
            possible_heads, \
                possible_tails = get_possible_heads_tails(self.kg_val,
                                                          possible_heads,
                                                          possible_tails)

        n_poss_heads = []
        n_poss_tails = []

        assert possible_heads.keys() == possible_tails.keys()

        for r in range(self.kg.n_rel):
            if r in possible_heads.keys():
                n_poss_heads.append(len(possible_heads[r]))
                n_poss_tails.append(len(possible_tails[r]))
                possible_heads[r] = list(possible_heads[r])
                possible_tails[r] = list(possible_tails[r])
            else:
                n_poss_heads.append(0)
                n_poss_tails.append(0)
                possible_heads[r] = list()
                possible_tails[r] = list()

        n_poss_heads = tensor(n_poss_heads)
        n_poss_tails = tensor(n_poss_tails)

        return possible_heads, possible_tails, n_poss_heads, n_poss_tails

    def corrupt_batch(self, heads, tails, relations, n_neg=None):
        """For each true triplet, produce a corrupted one not different from
        any other golden triplet. If `heads` and `tails` are cuda objects,
        then the returned tensors are on the GPU.

        Parameters
        ----------
        heads: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of heads of the relations in the
            current batch.
        tails: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of tails of the relations in the
            current batch.
        relations: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of relations in the current
            batch. This is optional here and mainly present because of the
            interface with other NegativeSampler objects.

        Returns
        -------
        neg_heads: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of negatively sampled heads of
            the relations in the current batch.
        neg_tails: torch.Tensor, dtype: torch.long, shape: (batch_size)
            Tensor containing the integer key of negatively sampled tails of
            the relations in the current batch.
        """
        device = heads.device
        assert (device == tails.device)

        batch_size = heads.shape[0]
        neg_heads, neg_tails = heads.clone(), tails.clone()

        # Randomly choose which samples will have head/tail corrupted
        mask = bernoulli(self.bern_probs[relations]).double()
        n_heads_corrupted = int(mask.sum().item())

        # Get the number of possible entities for head and tail
        n_poss_heads = self.n_poss_heads[relations[mask == 1]]
        n_poss_tails = self.n_poss_tails[relations[mask == 0]]

        assert n_poss_heads.shape[0] == n_heads_corrupted
        assert n_poss_tails.shape[0] == batch_size - n_heads_corrupted

        # Choose a rank of an entity in the list of possible entities
        choice_heads = (n_poss_heads.float() *
                        rand((n_heads_corrupted,))).floor().long()
        choice_tails = (n_poss_tails.float() *
                        rand((batch_size - n_heads_corrupted,))).floor().long()

        corr = []
        rels = relations[mask == 1]
        for i in range(n_heads_corrupted):
            r = rels[i].item()
            choices = self.possible_heads[r]
            if len(choices) == 0:
                # in this case the relation r has never been used with any head
                # choose one entity at random
                corr.append(randint(low=0, high=self.n_ent, size=(1,)).item())
            else:
                corr.append(choices[choice_heads[i].item()])
        neg_heads[mask == 1] = tensor(corr, device=device).long()

        corr = []
        rels = relations[mask == 0]
        for i in range(batch_size - n_heads_corrupted):
            r = rels[i].item()
            choices = self.possible_tails[r]
            if len(choices) == 0:
                # in this case the relation r has never been used with any tail
                # choose one entity at random
                corr.append(randint(low=0, high=self.n_ent, size=(1,)).item())
            else:
                corr.append(choices[choice_tails[i].item()])
        neg_tails[mask == 0] = tensor(corr, device=device).long()

        return neg_heads.long(), neg_tails.long()


def get_possible_heads_tails(kg, possible_heads=None, possible_tails=None):
    """Gets for each relation of the knowledge graph the possible heads and
    possible tails.

    Parameters
    ----------
    kg: `torchkge.data_structures.KnowledgeGraph`
    possible_heads: dict, optional (default=None)
    possible_tails: dict, optional (default=None)

    Returns
    -------
    possible_heads: dict, optional (default=None)
        keys: relation indices, values: set of possible heads for each
        relations.
    possible_tails: dict, optional (default=None)
        keys: relation indices, values: set of possible tails for each
        relations.

    """

    if possible_heads is None:
        possible_heads = defaultdict(set)
    else:
        assert type(possible_heads) == dict
        possible_heads = defaultdict(set, possible_heads)
    if possible_tails is None:
        possible_tails = defaultdict(set)
    else:
        assert type(possible_tails) == dict
        possible_tails = defaultdict(set, possible_tails)

    for i in range(kg.n_facts):
        possible_heads[kg.relations[i].item()].add(kg.head_idx[i].item())
        possible_tails[kg.relations[i].item()].add(kg.tail_idx[i].item())

    return dict(possible_heads), dict(possible_tails)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = ['torch>=1.2.0', 'tqdm', 'pandas']

setup_requirements = ['pytest-runner']

test_requirements = ['pytest']

setup(
    author="TorchKGE Developers",
    author_email='aboschin@enst.fr',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    description="Knowledge Graph embedding in Python and PyTorch.",
    license="BSD license",
    long_description=readme,
    include_package_data=True,
    keywords='torchkge',
    name='torchkge',
    url='https://github.com/torchkge-team/torchkge',
    packages=find_packages(),
    install_requires=requirements,
    setup_requires=setup_requirements,
    tests_require=test_requirements,
    test_suite='tests',
    version='0.16.20',
    zip_safe=False,
)



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from torch import tensor
from torch.nn import Embedding
from torch.nn.init import xavier_uniform_

import pickle
import tarfile

from .data import get_data_home

from os import makedirs, remove
from os.path import exists
from urllib.request import urlretrieve


def init_embedding(n_vectors, dim):
    """Create a torch.nn.Embedding object with `n_vectors` samples and `dim`
    dimensions. It is then initialized with Xavier uniform distribution.
    """
    entity_embeddings = Embedding(n_vectors, dim)
    xavier_uniform_(entity_embeddings.weight.data)

    return entity_embeddings


def load_embeddings(model, dim, dataset, data_home=None):

    if data_home is None:
        data_home = get_data_home()
    data_path = data_home + '/models/'
    targz_file = data_path + '{}_{}_{}.tar.gz'.format(model, dataset, dim)
    pkl_file = data_path + '{}_{}_{}.pkl'.format(model, dataset, dim)
    if not exists(pkl_file):
        if not exists(data_path):
            makedirs(data_path, exist_ok=True)
        urlretrieve("https://graphs.telecom-paris.fr/data/torchkge/models/{}_{}_{}.tar.gz".format(model, dataset, dim),
                    targz_file)
        with tarfile.open(targz_file, 'r') as tf:
            tf.extractall(data_path)
        remove(targz_file)

    with open(pkl_file, 'rb') as f:
        state_dict = pickle.load(f)

    return state_dict


def get_true_targets(dictionary, e_idx, r_idx, true_idx, i):
    """For a current index `i` of the batch, returns a tensor containing the
    indices of entities for which the triplet is an existing one (i.e. a true
    one under CWA).

    Parameters
    ----------
    dictionary: default dict
        Dictionary of keys (int, int) and values list of ints giving all
        possible entities for the (entity, relation) pair.
    e_idx: torch.Tensor, shape: (batch_size), dtype: torch.long
        Tensor containing the indices of entities.
    r_idx: torch.Tensor, shape: (batch_size), dtype: torch.long
        Tensor containing the indices of relations.
    true_idx: torch.Tensor, shape: (batch_size), dtype: torch.long
        Tensor containing the true entity for each sample.
    i: int
        Indicates which index of the batch is currently treated.

    Returns
    -------
    true_targets: torch.Tensor, shape: (batch_size), dtype: torch.long
        Tensor containing the indices of entities such that
        (e_idx[i], r_idx[i], true_target[any]) is a true fact.

    """
    true_targets = dictionary[e_idx[i].item(), r_idx[i].item()].copy()
    if len(true_targets) == 1:
        return None
    true_targets.remove(true_idx[i].item())
    return tensor(list(true_targets)).long()



import pandas as pd
import unittest

from torch import long

from torchkge.data_structures import KnowledgeGraph
from torchkge.evaluation import LinkPredictionEvaluator, TripletClassificationEvaluator
from torchkge.models import TransEModel


class TestUtils(unittest.TestCase):

    def setUp(self):
        df = pd.DataFrame([[0, 1, 0], [0, 2, 0], [0, 3, 0], [0, 4, 0], [1, 2, 1], [1, 3, 2], [2, 4, 0], [3, 4, 4],
                           [5, 4, 0]], columns=['from', 'to', 'rel'])
        self.kg = KnowledgeGraph(df)

    def checkSanityLinkPrediction(self, evaluator):
        assert evaluator.rank_true_heads.dtype == long
        assert evaluator.rank_true_tails.dtype == long
        assert evaluator.filt_rank_true_heads.dtype == long
        assert evaluator.filt_rank_true_tails.dtype == long

        assert evaluator.rank_true_heads.shape[0] == len(self.kg)
        assert evaluator.rank_true_tails.shape[0] == len(self.kg)
        assert evaluator.filt_rank_true_heads.shape[0] == len(self.kg)
        assert evaluator.filt_rank_true_tails.shape[0] == len(self.kg)

    def test_LinkPredictionEvaluator(self):
        model = TransEModel(100, self.kg.n_ent, self.kg.n_rel, 'L1')

        evaluator = LinkPredictionEvaluator(model, self.kg)
        self.checkSanityLinkPrediction(evaluator)

        evaluator.evaluate(b_size=len(self.kg), k_max=10)
        self.checkSanityLinkPrediction(evaluator)

    def test_TripletClassificationEvaluator(self):
        model = TransEModel(100, self.kg.n_ent, self.kg.n_rel, 'L1')
        kg1, kg2 = self.kg.split_kg(sizes=(4, 5))
        # kg2 contains all relations so it will be used as validation
        evaluator = TripletClassificationEvaluator(model, kg2, kg1)
        assert evaluator.thresholds is None
        assert not evaluator.evaluated

        evaluator.evaluate(b_size=len(self.kg))
        assert evaluator.evaluated
        assert evaluator.thresholds is not None
        assert (len(evaluator.thresholds.shape) == 1) & (evaluator.thresholds.shape[0] == self.kg.n_rel)



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""


class NotYetEvaluatedError(Exception):
    def __init__(self, message):
        super().__init__(message)


class SizeMismatchError(Exception):
    def __init__(self, message):
        super().__init__(message)


class WrongDimensionError(Exception):
    def __init__(self, message):
        super().__init__(message)


class NotYetImplementedError(Exception):
    def __init__(self, message):
        super().__init__(message)


class WrongArgumentsError(Exception):
    def __init__(self, message):
        super().__init__(message)


class SanityError(Exception):
    def __init__(self, message):
        super().__init__(message)


class SplitabilityError(Exception):
    def __init__(self, message):
        super().__init__(message)


class NoPreTrainedVersionError(Exception):
    def __init__(self, message):
        super().__init__(message)
